/**
 * Author: Valerio Schiavoni <valerio.schiavoni@gmail.com>
 */
package org.objectweb.fractal.bf.connectors.rmi;

import java.rmi.Remote;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.bf.AbstractBindingFactoryPlugin;
import org.objectweb.fractal.bf.BindHints;
import org.objectweb.fractal.bf.BindingFactoryException;
import org.objectweb.fractal.bf.SkeletonGenerationException;
import org.objectweb.fractal.bf.proxies.StubComponentFactory;
import org.objectweb.fractal.util.Fractal;

/**
 * 
 */
public class RmiConnector extends
		AbstractBindingFactoryPlugin<RmiExportHints, RmiBindHints> {

	Logger log = Logger.getLogger(RmiConnector.class.getCanonicalName());

	public static Remote skeletonRef;

	/**
	 * here to please Julia.
	 */
	public RmiConnector() {
		this.initialHints = new HashMap<String, String>();
	}

	public RmiConnector(Map<String, String> hints) {
		this.initialHints = hints;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#finalizeSkeleton(org.objectweb.fractal.api.Component,
	 *      java.util.Map)
	 */
	public void finalizeSkeleton(Component skel,
			Map<String, String> finalizationHints)
			throws BindingFactoryException {

		log.warning("FinalizationHints map currently ignored");

		try {
			RmiConnectorAttributes rca = (RmiConnectorAttributes) skel
					.getFcInterface(RmiConnectorAttributes.RMI_CONNECTOR_ATTRIBUTES);

			rca.startRmi();

		} catch (NoSuchInterfaceException e) {
			throw new BindingFactoryException(e);
		}

	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getStubComponentFactory()
	 */
	public StubComponentFactory getStubComponentFactory() {
		return new RmiConnectorStubComponentFactoryImpl();
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#createSkel(java.lang.String,
	 *      java.lang.Object, org.objectweb.fractal.api.Component,
	 *      org.objectweb.fractal.bf.ExportHints)
	 */
	public Component createSkel(String interfaceName, Object serverItf,
			Component owner, RmiExportHints exportHints)
			throws SkeletonGenerationException {
		log.info("RmiConnector - Creating skeleton.");

		String name = exportHints.getServiceName();
		String host = exportHints.getHost();
		String port = exportHints.getPort();

		log.info("Service Name: " + name + " host:" + host + " port:" + port);

		// a fractal component whose content is a rmi skeleton.
		Component compRmiSkel = null;

		try {

			Factory adlFactory = FactoryFactory
					.getFactory(FactoryFactory.FRACTAL_BACKEND);

			compRmiSkel = (Component) adlFactory.newComponent(
					"org.objectweb.fractal.bf.connectors.rmi.RmiServerBridge",
					null);

			// start the skeleton
			Fractal.getLifeCycleController(compRmiSkel).startFc();

			RmiConnectorAttributes rmiServiceAttributes = (RmiConnectorAttributes) compRmiSkel
					.getFcInterface(RmiConnectorAttributes.RMI_CONNECTOR_ATTRIBUTES);
			rmiServiceAttributes.setServiceName(name);
			rmiServiceAttributes.setHost(host);
			rmiServiceAttributes.setPort(port);
			/*
			 * this should not be necessary if the ADL factory could take into
			 * account instance of parameter objects
			 */
			rmiServiceAttributes.setServiceObject(serverItf);

		} catch (Exception e) {

			e.printStackTrace();
			throw new SkeletonGenerationException(
					"Some error occurred while exporting using the JavaRMI plugin ",
					e);
		}

		return compRmiSkel;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getExportHints()
	 */
	public RmiExportHints getExportHints() {
		RmiExportHints rmiExportHints = new RmiExportHints();

		log.fine("Initial hints: " + this.initialHints.toString());

		String registryHost = computeRegistryHost();
		rmiExportHints.setHost(registryHost);

		String registryPort = computeRegistryPort();
		rmiExportHints.setPort(registryPort);

		String serviceName = computeServiceName();
		rmiExportHints.setServiceName(serviceName);
		return rmiExportHints;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getBindHints()
	 */
	public BindHints getBindHints() {
		RmiBindHints rmiBindHints = new RmiBindHints();
		String registryHost = computeRegistryHost();
		rmiBindHints.setHost(registryHost);

		String registryPort = computeRegistryPort();
		rmiBindHints.setPort(registryPort);

		String serviceName = computeServiceName();
		rmiBindHints.setServiceName(serviceName);
		return rmiBindHints;
	}

	/**
	 * @return
	 */
	private String computeServiceName() {
		String serviceName = initialHints
				.get(RmiConnectorConstants.SERVICE_NAME);
		log.finer("Service name: " + serviceName);
		return serviceName;
	}

	/**
	 * @return
	 */
	private String computeRegistryPort() {
		String registryPort = initialHints
				.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_PORT);
		if (registryPort == null) {
			log
					.warning("You didn't specify the port for the RMI registry, the default port [1099] will be used");
			registryPort = "1099";
		}
		return registryPort;
	}

	/**
	 * @return
	 */
	private String computeRegistryHost() {
		String registryHost = initialHints
				.get(RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS);
		if (registryHost == null) {
			log
					.warning("You didn't specify "
							+ RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS
							+ ", will use '"
							+ RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS_DEFAULT
							+ "' ");
			registryHost = RmiConnectorConstants.JAVA_RMI_REGISTRY_HOST_ADDRESS_DEFAULT;
		}
		return registryHost;
	}

	/**
	 * @see org.objectweb.fractal.bf.BindingFactoryPlugin#getPluginIdentifier()
	 */
	public String getPluginIdentifier() {
		return "rmi";
	}
}
