/**
 * Author: Valerio Schiavoni <valerio.schiavoni@gmail.com>
 */
package org.objectweb.fractal.bf.connectors.rmi;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.rmi.Remote;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.logging.Logger;

import org.objectweb.fractal.bf.BindingFactoryException;

/**
 * Dispatch calls to the RMI registry returning the result to the stub component
 * this handler will be part of.
 */
public class RmiInvocationHandler implements InvocationHandler {

	Logger log = Logger
			.getLogger(RmiInvocationHandler.class.getCanonicalName());

	/**
	 * the name of the service this invocation handler will dispatch calls to.
	 */
	private final String name;
	/**
	 * The host address of the RMI registry.
	 */
	private final String hostAddress;
	/**
	 * The port of the RMI regitry.
	 */
	private final String registryPort;

	Registry registry;

	/**
	 * @param params
	 */
	public RmiInvocationHandler(String name, String hostAddress,
			String registryPort) {
		this.name = name;
		this.hostAddress = hostAddress;
		this.registryPort = registryPort;
	}

	public RmiInvocationHandler(RmiBindHints bindHints) {
		this(bindHints.getServiceName(), bindHints.getHost(), bindHints
				.getPort());
	}

	/**
	 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object,
	 *      java.lang.reflect.Method, java.lang.Object[])
	 */
	public Object invoke(Object proxy, Method method, Object[] args)
			throws Throwable {
		log.config("Looking for JavaRMI registry at " + hostAddress + ":"
				+ registryPort);
		if (registry == null) {
			log.info("RmiInvocationHandler: JavaRMI registry found: "
					+ registry);
			this.registry = LocateRegistry.getRegistry(hostAddress, Integer
					.valueOf(registryPort));
			if (registry == null) {
				throw new BindingFactoryException(
						"Cannot locate JavaRMi registry");
			}
		}

		Remote remote = registry.lookup(this.name);

		log.fine("Remote reference for service recovered from registry: "
				+ remote);

		final Method remoteMethod = remote.getClass().getMethod(
				method.getName(), method.getParameterTypes());

		log.finer("Method to be called on remote reference: "
				+ remoteMethod.toString());

		Object result = remoteMethod.invoke(remote, args);

		log.fine("Result returned by remote reference: " + result);

		return result;
	}
}
